import pozo.renderers as pzr
from IPython.core.display import HTML
import html

class TreeTable(pzr.Renderer):
    def _get_basic(self, item):
        name = item.get_name() # dict
        theme = item.get_theme() # dict
        theme_t = type(theme)
        theme_context = theme.get_context() # dict
        return {
                "name": name,
                "theme": theme,
                "theme_type": theme_t,
                "theme_context": theme_context,
                "children": [],
                }

    def _get_graph_dict(self, graph):
        graph_node = self._get_basic(graph)
        return graph_node

    def _get_track_dict(self, track):
        track_node = self._get_basic(track)
        return track_node

    def _get_axis_dict(self, axis):
        axis_node = self._get_basic(axis)
        return axis_node

    def _get_trace_dict(self, trace):
        mnemonic = trace.get_mnemonic()
        values_t = type(trace.get_data())
        depth_t = type(trace.get_depth())
        length = len(trace.get_data())
        trace_dict = self._get_basic(trace)
        trace_dict["mnemonic"] = mnemonic
        trace_dict["data_type"] = values_t
        trace_dict["depth_type"] = depth_t
        trace_dict["data_unit"] = trace.get_unit()
        trace_dict["depth_unit"] = trace.get_depth_unit()
        trace_dict["length"] = length
        return trace_dict

    def render(self, graph, **kwargs):
        graph_node = self._get_graph_dict(graph)
        total_traces = 0
        for track in graph:
            track_node = self._get_track_dict(track)
            for axis in track:
                axis_node = self._get_axis_dict(axis)
                for trace in axis:
                    trace_node = self._get_trace_dict(trace)
                    axis_node["children"].append(trace_node)
                    total_traces += 1
                track_node["children"].append(axis_node)
            graph_node["children"].append(track_node)


        style = "\"border: 1px solid gray; text-align:center\""
        output = "<table>"

        output += "<tr>"
        output += f"<td colspan=\"{total_traces}\" style={style}>"
        output += "<h3>Graph</h3>"

        tracks = []
        for key, value in graph_node.items():
            if key != "children":
                output += f"<p style=\"text-align:center\">{html.escape(str(key))}: {html.escape(str(value))}</p>"
            else:
                tracks = value # tracks is just a list of tracks
        output += "</td></tr>"

        output += "<tr>"
        axes = []
        for track in tracks:
            colspan = 0
            for axis in track['children']:
                colspan += len(axis['children'])
            output += f"<td colspan=\"{colspan}\" style={style}>"
            output += "<h6>Track</h6>"
            for key, value in track.items():
                if key != "children":
                    output += f"<p style=\"text-align:center\">{html.escape(str(key))}: {html.escape(str(value))}</p>"
                else:
                    axes.append(value) # axes is a list of lists of axes
            output += "</td>"
        output += "</tr>"

        output += "<tr>"
        traces = []
        for axes_list in axes:
            for axis in axes_list:
                output += f"<td colspan=\"{len(axis['children'])}\" style={style}>"
                output += "<h6>Axis</h6>"
                for key, value in axis.items():
                    if key != "children":
                        output += f"<p style=\"text-align:center\">{html.escape(str(key))}: {html.escape(str(value))}</p>"
                    else:
                        traces.append(value)
                output += "</td>"
        output += "</tr>"

        for traces_list in traces:
            for trace in traces_list:
                output += f"<td style={style}>"
                output += "<h6>Data</h6>"
                for key, value in trace.items():
                    if key != "children":
                        output += f"<p style=\"text-align:center\">{html.escape(str(key))}: {html.escape(str(value))}</p>"
                    else:
                        pass
                output += "</td>"
        output += "</tr>"
        output += "</table>"
        display(HTML(output))





class Renderer():
    pass

# These imports we want in this name space, but they use the stuff above!
from .plotly import Plotly, CrossPlot # this is not great namespacing TODO
from .tree_table import TreeTable


import copy
import os
import math
import warnings
import re
import weakref
import multiprocessing
from ipywidgets import IntProgress

import numpy as np
from IPython.display import Javascript, display # Part of Hack #1
import plotly as plotly
import plotly.graph_objects as go
import pint

import pozo
import pozo.renderers as pzr
import pozo.themes as pzt
import pozo.units as pzu

re_space = re.compile(' ')
re_power = re.compile(r'\*\*')

def javascript():
    add_scroll = '''var css = '.plot-container { overflow: auto; }',
head = document.getElementsByTagName('head')[0],
style = document.createElement('style');

style.type = 'text/css';
if (style.styleSheet){
  style.styleSheet.cssText = css;
} else {
  style.appendChild(document.createTextNode(css));
}
console.log("Executed javascript()")
head.appendChild(style);'''
    display(Javascript(add_scroll))

javascript()

# TODO FOR CROSSPLOT MUST X Y AND ALL COLOR HAVE THE SAME SHAPE OR ELSE BREAK INTERACTIVITY

# You can create your own dictionary like 'defaults' and
# construct a Style() with it (pass it in as `template`)
# Specifying values that are marked as "generated" will
# lead to undefined behavior, probably be overwritten.
# It may be better to inherit the Style() class and
# define your own class that overrides the generators.
# Furthermore, certain keywords are marked as required,
# since the generators depend on them and expect them.

defaults = dict(
    # Non-Plotly style values, used to generate Plotly layout values. These are all "required".
    track_margin = 15,      # margin between each track
    depth_axis_width = 35,  # size of depth axis if it's not the first thing, will also have margin above
    axis_label_height = 80, # size of each xaxis when stacked
    y_annotation_gutter = 10, # space for writing text

    # Plotly style values, structured like a plotly layout dictionary.
    # Important to know that `xaxes_template` will be used to create several `xaxis1`,`xaxis2`, `xaxisN` keys.
    # `xaxis_template` will not make it into the final layout dictionary passed to Plotly.
    plotly = dict(
        height = 900,          # required
        showlegend = True,
        legend = dict(
            yanchor = "top",
            y = 1,
            xanchor = "left",
#           x = 0,             # generated
        ),
        margin = dict(l=15, r=15, t=5, b=5),
        plot_bgcolor = '#FFFFFF',
        paper_bgcolor = '#FFFFFF',
#       width=?                # generated

        yaxis1 = dict(
                visible=True, # Would hdie gridelines too, so hide ticks and labels instead of axis.
                showgrid=True,
                zeroline=False,
                showline=False,
                position=0,
                gridcolor="#f0f0f0",
#               domain=[?,?],  # generated
#               maxallowed=,   # generated
#               minallowed=,   # generated
#               range=[?,?],   # generated
        ),

        xaxis_template = dict(
            showgrid=False, # TODO this is tough, grid where?
            zeroline=False,
            gridcolor="#f0f0f0", # this needs to be generated
            showline=True,
            linewidth=2,
            tickmode= 'auto',
            nticks= 0,
            ticks="outside",
            tickwidth=1,
            ticklen=6,
            tickangle=55,
#           side=['']              # generated
#           tickcolor=='#???',     # generated
#           linecolor='#???',      # generated
#           title=dict(
#               text="???",        # generated
#               font=dict(
#                   color="#???"   # generated
#           ),
#           tickfont=dict(
#               color="#???"       # generated
#           )
#           domain = [?,?]         # generated
#           anchor=""              # generated
#           autoshift=             # Doesn't work for x-axes!
#           position=              # generated
#           overlaying=            # generated
        ),
    ),
)
# different hover breaks sometimes
# different date stuff break sometimes (wasn't showing me august?)

default_hovertemplate = 'Depth: %{y}, Value: %{x}' # how to get this to be the name

GRAPH_HEIGHT_MIN = 125

# move this?
def is_array(value):
    if isinstance(value, str): return False
    if hasattr(value, "magnitude"):
        return is_array(value.magnitude)
    return hasattr(value, "__len__")

def toTarget(axis):
    return axis[0] + axis[-1]

# So this renderer will have a list of tracks, and their contained axis, but not the data
# Its all position data, not styling data
class Plotly(pzr.Renderer):
    def __init__(self, template=defaults):
        javascript() # double it up
        self.template = copy.deepcopy(template)
        self.xaxis_template = copy.deepcopy(self.template["plotly"]["xaxis_template"])
        del self.template["plotly"]["xaxis_template"]
        self.last_fig = None

    def _hidden(self, themes):
        hidden = themes["hidden"]
        if hidden: themes.pop()
        return hidden

   # TODO, units (find nearest with units too)
    def _process_graph_note(self, note, x0, x1, yref):
        if 'range' not in note:
            raise ValueError("Depth note seems to be missing range")
        # TODO: check numerics
        type = 'line'
        if isinstance(note['range'], tuple): type='rect'
        shape = None
        if type=='line':
            shape = dict(
                type='line',
                xref='paper',
                x0=x0,
                x1=x1,
                yref=yref,
                y0=note['range'],
                y1=note['range'],
                line_width=note['weight'] if 'weight' in note else 1,
                line_dash=note['style'] if 'style' in note else 'dot',
                line_color=note['color'] if 'color' in note else 'black',
            )
        else:
            shape = dict(
                type='rect',
                xref='paper',
                x0=x0,
                x1=x1,
                yref=yref,
                y0=note['range'][0],
                y1=note['range'][1],
                line_width=0,
                fillcolor=note['color'] if 'color' in note else 'lightskyblue',
                layer="below",
                opacity=.5,
            )
        annotation = None
        if not ('show_text' in note and not note['show_text']):
            annotation = dict(
                text=note['text'] if 'text' in note else "",
                xref="paper",
                x=1,
                yref=yref,
                y=note['range'] if type=='line' else note['range'][0],
                yshift=5,
                showarrow=False,
            )
        return shape, annotation

    def get_layout(self, graph, xp=None, **kwargs):
        if not isinstance(graph, pozo.Graph):
            raise TypeError("Layout must be supplied a graph object.")
        if not len(graph):
            raise ValueError("There are no tracks, there is nothing to lay out.")
        show_depth = kwargs.pop("show_depth", True)
        depth_axis_position = kwargs.pop("depth_position", 0)
        override_theme = kwargs.pop("override_theme", None)
        override_theme = kwargs.pop("theme_override", override_theme)
        height = kwargs.get("height", None)
        depth_range = kwargs.get("depth", None)
        if depth_range is not None and ( not isinstance(depth_range, (tuple, list)) or len(depth_range) != 2 ):
            raise TypeError(f"Depth range must be a list or tuple of length two, not {depth_range}")

        layout = copy.deepcopy(self.template["plotly"])
        if height is not None:
            layout["height"] = height
        if layout["height"] < GRAPH_HEIGHT_MIN:
            raise ValueError("125px is the minimum total height")

        tracks_y_axis = "yaxis1"
        xp_y_axis = None
        xp_x_axis = None
        if xp is not None:
            tracks_y_axis = "yaxis2"
            xp_y_axis = "yaxis1"
            xp_x_axis = "xaxis1"
            layout[tracks_y_axis] = copy.deepcopy(layout['yaxis1'])
            del layout['yaxis1']
            # Honestly all communication between renderers should maybe not happen inside any get_layout
            # nor any get_trace

        # TODO constrain positions and domains

        # posmap is an object eventual assigned to the figure containing axis/graph/position data.
        # it's used to pull out reference numbers easily
        posmap = {
            'xp_x': xp_x_axis,
            'xp_y': xp_y_axis,
            'track_y': tracks_y_axis,
            'pixel_height': layout['height'],
            'pixel_width': 4, # start drawing axes at 4 makes it look a little nicer on the left
            'pixel_cursor': 4,
            'with_xp': xp is not None,
            'xp_end': 0,
            'show_depth': show_depth,
            'depth_track_number': depth_axis_position if show_depth else 0, # putting at 0 easier if turned off
            'depth_auto_left': not depth_axis_position, # weird because if `width_xp` this should be false, but logic works
            'depth_auto_right': False, # calc later
            'tracks_axis_numbers': [],
            'tracks_pixel_widths': [],
            'tracks_x_domains': [],
            'tracks_y_domain': None,
            'axis_line_positions': [],
            'total_axes': int(xp is not None),
            'y_bottom': self.template['y_annotation_gutter']/layout['height'],
            }


        # we do a lot of Y calculations in the first run
        max_axis_stack = 0
        track_index = -1
        for track in graph.get_tracks():
            track_index += 1
            if pzt.ThemeStack(track.get_theme())['hidden'] or len(track) == 0: continue
            if track_index and posmap['depth_track_number'] == track_index: posmap['tracks_axis_numbers'].append("depth")
            posmap['tracks_axis_numbers'].append([])
            for axis in track.get_axes():
                if pzt.ThemeStack(track.get_theme())['hidden']: continue
                posmap['total_axes'] += 1 # plotly axis indicies are base 1 so increment first
                posmap['tracks_axis_numbers'][-1].append(posmap['total_axes'])
            max_axis_stack = max(len(posmap['tracks_axis_numbers'][-1]), max_axis_stack)
        if not max_axis_stack: raise ValueError("Didn't find an axes, not going to render")
        if max_axis_stack == 1: posmap['axis_line_positions'] = [1]
        elif max_axis_stack > 1:
            for i in reversed(range(max_axis_stack+1)):
                posmap['axis_line_positions'].append(
                    1 - i * (self.template["axis_label_height"]/posmap['pixel_height'])
                    )
        layout[tracks_y_axis]["domain"] = posmap['tracks_y_domain'] = ( posmap['y_bottom'], posmap['axis_line_positions'][0] )
        if posmap['depth_track_number'] >= len(posmap['tracks_axis_numbers']):
            posmap['depth_auto_right'] = True
        max_text = 0
        for name, note in graph.depth_notes.items():
            max_text = max(max_text, len(note['text'] if 'text' in note else name))
        posmap['x_annotation_pixel_width'] = max_text*10
        posmap['pixel_width'] += max_text*10

        ## in the second pass, we apply units and styles, and information from the first run
        axes_styles = [] # why not declare axes_styles earlier and then apply axes styles from the start, instead of here

        ymin = float('inf')
        ymax = float('-inf')
        y_unit = None

        themes = pzt.ThemeStack(pzt.default_theme, theme = override_theme)
        themes.append(graph.get_theme())
        if self._hidden(themes): return {}

        track_index = -1 # we can't use enumerate() becuase sometimes we skip iterations in the loop
        for track in graph.get_tracks():
            themes.append(track.get_theme())
            if self._hidden(themes) or len(track) == 0: continue
            track_index += 1
            if track_index and posmap['tracks_axis_numbers'][track_index] == 'depth': # don't do it for 0
                track_index +=1
                posmap['tracks_pixel_widths'].append(self.template['depth_axis_width'])


            axis_index = -1 # same story with track_index, need to skip iterations, enumerate() won't work
            for axis in track.get_axes():
                themes.append(axis.get_theme())
                if self._hidden(themes): continue
                axis_index += 1

                if themes["range_unit"] is not None:
                    range_unit = pzu.registry.parse_units(themes["range_unit"])
                else:
                    range_unit = None
                data_unit = None
                for trace in axis:
                    themes.append(trace.get_theme())
                    if self._hidden(themes): continue

                    if data_unit is not None and data_unit != trace.get_unit():
                        raise ValueError(f"Data being displayed on one axis must be exactly the same unit. {data_unit} is not {trace.get_unit()}")
                    else:
                        data_unit = trace.get_unit()
                        if not (data_unit is None or range_unit is None or range_unit.is_compatible_with(data_unit)):
                            raise pint.DimensionalityError(range_unit, data_unit, extra_msg="range_unit set by theme is not compatible with data units")
                    themes.pop()
                    if y_unit is not None and trace.get_depth_unit() is not None:
                        if y_unit != trace.get_depth_unit():
                            raise ValueError(f"All depth axis must have the same unit. You must transform the data. {y_unit} is not {trace.get_depth_unit()}")
                    y_unit = trace.get_depth_unit()
                    ymin = min(pzu.Q(trace.get_depth()[0], y_unit), pzu.Q(ymin, y_unit)).magnitude
                    ymax = max(pzu.Q(trace.get_depth()[-1], y_unit), pzu.Q(ymax, y_unit)).magnitude

                if data_unit is None: data_unit = range_unit
                if range_unit is None: range_unit = data_unit # both None, or both whatever wasn't None
                if data_unit != range_unit:
                    xrange = pzu.Q(themes["range"], range_unit).m_as(data_unit)
                else:
                    xrange = themes["range"]
                # So we've just created xrange which is the data_unit
                # But here we'd want to override ticks and such to the range unit
                color = themes["color"]

                scale_type = themes["scale"]
                axis_style = dict(
                    **self.xaxis_template
                )
                append_unit = ""
                if data_unit is not None:
                    unit = format(data_unit, '~')
                    unit = re_space.sub('', unit)
                    unit = re_power.sub('^', unit)
                    append_unit = " (" + unit + ")"
                axis_style['title'] = dict(text=axis.get_name() + append_unit, font=dict(color=color), standoff=0,)
                axis_style['linecolor'] = color
                axis_style['tickcolor'] = color
                axis_style['tickfont']  = dict(color=color,)

                if axis_style is not None:
                    axis_style['type'] = scale_type
                    if scale_type == "log":
                        xrange = [math.log(xrange[0], 10), math.log(xrange[1], 10)]
                if xrange is not None:
                    axis_style['range'] = xrange

                axis_style['side'] = "top"
                if axis_index: # is a secondary axis
                    axis_style['anchor'] = "free" # TODO we might be able to just change this, check
                    axis_style['overlaying'] = "x" + str(posmap['tracks_axis_numbers'][track_index][0])
                # TODO why do we need position on automatically anchored axis in the case of XP? (from previous comment)
                axis_style['position'] = min(posmap['axis_line_positions'][axis_index], 1)
                axes_styles.append(axis_style)
                themes.pop()
            posmap['pixel_width'] += themes["track_width"] + self.template["track_margin"]
            posmap['tracks_pixel_widths'].append(themes["track_width"])
            themes.pop()
        posmap['pixel_width'] -= self.template["track_margin"] # we don't need margin at the end
        if posmap['depth_track_number'] > 0: posmap['pixel_width'] += self.template['depth_axis_width'] # We'll need space

        # all stuff dependent on position could go here
        if xp is not None:
            posmap["pixel_width"] += layout["height"] # create a square for the crossplot
            if posmap['depth_auto_left']:
                posmap["pixel_width"] += self.template['depth_axis_width']
            posmap['xp_end'] = layout["height"] / (posmap["pixel_width"])
            xp_layout = xp.create_layout(container_width = posmap["pixel_width"], size = layout["height"]) # pre posmap
            layout[xp_y_axis] = xp_layout["yaxis1"]
            layout[xp_x_axis] = xp_layout["xaxis1"]
            layout[xp_x_axis]["domain"] = (0, posmap['xp_end'])
            posmap["xp_domain"] = layout[xp_x_axis]["domain"]
            layout["legend"]["x"] = posmap['xp_end'] * .8
            posmap['pixel_cursor'] += layout["height"]
            if posmap['depth_auto_left']: posmap['pixel_cursor'] += self.template['depth_axis_width']
        track = 0
        for i, axis in enumerate(axes_styles):
            if 'overlaying' in axis:
                axis['domain'] = axes_styles[i-1]['domain']
            else:
                axis['domain'] = (posmap['pixel_cursor']/posmap['pixel_width'],
                                  (posmap['pixel_cursor']+posmap['tracks_pixel_widths'][track])/posmap['pixel_width'])
                posmap['pixel_cursor'] += posmap['tracks_pixel_widths'][track]+self.template["track_margin"]
                posmap['tracks_x_domains'].append(axis['domain'])
                track+= 1
                if ( track < len(posmap['tracks_axis_numbers']) and
                    posmap['tracks_axis_numbers'][track] == 'depth'
                ): # never checks track 0, that's auto
                    track+=1
                    layout[tracks_y_axis]['position'] = (-4 + posmap['pixel_cursor'] + self.template['depth_axis_width'])/posmap['pixel_width']
                    posmap['tracks_x_domains'].append((posmap['pixel_cursor']/posmap['pixel_width'], layout[tracks_y_axis]['position']))
                    posmap['pixel_cursor'] += self.template['depth_axis_width']
            layout["xaxis" + str(i + 1 + int(xp is not None))] = axis
        if not show_depth:
            layout[tracks_y_axis]['showticklabels'] = False
            layout[tracks_y_axis]['ticklen'] = 0
            layout[tracks_y_axis]['ticks'] = ""
        elif posmap['depth_auto_right']:
            layout[tracks_y_axis]['position'] = 1 - posmap['x_annotation_pixel_width']/posmap['pixel_width']
        elif posmap['depth_auto_left'] and xp is not None:
            layout[tracks_y_axis]['position'] = posmap['xp_end'] + self.template['depth_axis_width']/posmap['pixel_width']
        layout[tracks_y_axis]['maxallowed'] = ymax
        layout[tracks_y_axis]['minallowed'] = ymin
        if depth_range is not None:
            layout[tracks_y_axis]['range'] = [depth_range[1], depth_range[0]]
        else:
            layout[tracks_y_axis]['range'] = [ymax, ymin]
        layout['width']=posmap['pixel_width']
        self._last_posmap = posmap


        # TODO: add verts and track depth/ranges, it could be condensed into just passing posmap
        depth_margin = self.template['depth_axis_width']/posmap['pixel_width'] if (posmap['depth_auto_left'] and posmap['with_xp']) else 0
        layout['shapes'] = []
        layout['annotations'] = []
        for name, note in graph.depth_notes.items():
            if 'text' not in note: note['text']=name
            s, a = self._process_graph_note(note, posmap['xp_end']+depth_margin, 1, toTarget(posmap['track_y']))
            if s: layout['shapes'].append(s)
            if a: layout['annotations'].append(a)
        return layout

    def get_traces(self, graph, yaxis='y1', xstart=1, **kwargs):
        override_theme = kwargs.pop("override_theme", None)
        override_theme = kwargs.pop("theme_override", override_theme)
        traces = []
        num_axes = xstart
        themes = pzt.ThemeStack(pzt.default_theme, theme = override_theme)
        themes.append(graph.get_theme())
        if self._hidden(themes): return []
        for track in graph:
            themes.append(track.get_theme())
            if self._hidden(themes): continue
            for axis in track:
                themes.append(axis.get_theme())
                if self._hidden(themes): continue
                all_traces = []
                for trace in axis:
                    themes.append(trace.get_theme())
                    if self._hidden(themes): continue
                    color = themes["color"]
                    with warnings.catch_warnings():
                        warnings.filterwarnings(action='ignore', category=pint.UnitStrippedWarning, append=True)
                        all_traces.append(go.Scattergl(
                            x=trace.get_data(),
                            y=trace.get_depth(),
                            mode='lines', # nope, based on data w/ default
                            line=dict(color=color, width=1), # needs to be better, based on data
                            xaxis='x' + str(num_axes),
                            yaxis=toTarget(yaxis),
                            name = trace.get_name(),
                            hovertemplate = default_hovertemplate,
                            showlegend = False,
                        ))
                    themes.pop()
                num_axes += 1
                traces.extend(all_traces)
                themes.pop()
            themes.pop()
        return traces

    def render(self, graph, static=False, depth=None, xp=None, **kwargs):
        xp_depth = kwargs.pop("xp_depth", None)
        xp_depth_by_index = kwargs.pop("xp_depth_by_index", None)
        depth_lock = False
        if xp_depth or xp_depth_by_index:
            depth_lock = True

        by_index = False
        if xp_depth_by_index:
            if xp_depth: raise ValueError("Specify xp_depth or xp_depth_by_index")
            xp_depth = xp_depth_by_index
            by_index = True
        elif not xp_depth:
            xp_depth = depth

        color_lock = kwargs.pop("color_lock", {})
        # kwargs: theme_override, override_theme (same thing)

        # this generates XP layout too, I don't love, would rather set axes
        # separating XP from here would be good but let's call it techdebt
        layout = self.get_layout(graph, xp=xp, depth=depth, **kwargs)

        # what arguments do we need here now
        traces = self.get_traces(graph, xstart=2 if xp else 1, yaxis='yaxis2' if xp else 'yaxis1', **kwargs)
        if xp is not None:
            traces.extend(
                    xp.create_traces(
                        container_width=layout["width"],
                        depth_range=xp_depth,
                        by_index=by_index,
                        size = layout["height"],
                        static=static,
                        yaxis='y1',
                        color_lock=color_lock
                        )
                    )

        if static:
            self.last_fig = go.Figure(data=traces, layout=layout)
            return self.last_fig
        elif xp is None:
            self.last_fig = xpFigureWidget(data=traces, layout=layout)
            self.last_fig._lead_axis = "yaxis1"
        else:
            self.last_fig = xpFigureWidget(data=traces, layout=layout, renderer=xp, depth_lock=depth_lock)
            self.last_fig._lead_axis = "yaxis2"
            self.last_fig.link_depth_to(self.last_fig)
            xp.add_figure(self.last_fig)
        javascript() # double it up
        return self.last_fig

# Use this for all FigureWidgets, have way to return FigureWidget
class xpFigureWidget(go.FigureWidget):
    # pass the controlling graph here, and it will add a callback
    def link_depth_to(self, fig):
        if not isinstance(fig, go.FigureWidget):
            raise TypeError("Supplied fig argument bust be a go.FigureWidget ot have access to interactivity")
        fig.layout.on_change(self._depth_change_cb, fig._lead_axis+'.range', append=True)

    def __init__(self, data=None, layout=None, frames=None, skip_invalid=False, depth_lock=False, **kwargs):
        self._xp_renderer = kwargs.pop("renderer", None)
        if self._xp_renderer:
            self._depth_lock = True
        super().__init__(data=data, layout=layout, frames=frames, skip_invalid=skip_invalid, **kwargs)

    def _depth_change_cb(self, layout, new_range):
        if not self._depth_lock:
            self.set_depth_range(new_range)
        self._tracks_depth_range = new_range

    def lock_depth_range(self):
        if not self._xp_renderer: raise TypeError("lock_depth_range only applies to graphs with a crossplot")
        self._depth_lock = True

    def unlock_depth_range(self):
        if not self._xp_renderer: raise TypeError("unlock_depth_range only applies to graphs with a crossplot")
        self._depth_lock = False
        self.match_depth_range()

    def match_depth_range(self):
        if not self._xp_renderer: raise TypeError("match_depth_range only applies to graphs with a crossplot")
        self.set_depth_range(depth_range=self._tracks_depth_range)

    def set_depth_range(self, depth_range=None):
        if not self._xp_renderer: raise TypeError("set_depth_range only applies to graphs with a crossplot")
        depth_range = sorted(depth_range) if depth_range else None
        color_range_queue=[]
        with self.batch_update():
            for trace in self['data']:
                if not trace.meta or trace.meta.get("filter", None) != 'depth': continue
                trace.x = self._xp_renderer.x.get_data(slice_by_depth=depth_range)
                trace.y = self._xp_renderer.y.get_data(slice_by_depth=depth_range)
                color_data = trace.meta.get('color_data', None)
                if color_data == 'depth': # if xp were to change x or y, this would be wrong in renderer
                    trace.marker.color = self._xp_renderer.x.get_depth(slice_by_depth=(depth_range))
                elif color_data in self._xp_renderer._colors_by_id:
                    trace.marker.color = self._xp_renderer._colors_by_id[color_data].get_data(slice_by_depth=depth_range)
                if 'color_range' in trace.meta and trace.meta['color_range'] is not None: # no need for (None)?
                    color_range_queue.append((trace.name, trace.meta['color_range']))
        with self.batch_update(): # instead of two batches, color_range could take the new info as arguments
            for item in color_range_queue:
                self.set_color_range(item[0], color_range=item[1])

    # we have to set self._color_range so that can recalculate it when depth changes
    def set_color_range(self, name, color_range=(None), auto=False, lock=False):
        if not self._xp_renderer: raise TypeError("set_color_range only applies to graphs with a crossplot")
        for trace in self['data']:
            if trace.meta and 'color_data' in trace.meta:
                if trace.name != name: continue
                if color_range: color_range = list(sorted(color_range))
                calc_key = 'colorscale_calculated'
                select_key = 'colorscale_selected'

                # if we never calculated or what we have is different than what we calculaed
                if (calc_key not in trace.meta
                    or trace.meta[calc_key] != trace.marker.colorscale
                    ):
                    trace.meta[select_key] = trace.marker.colorscale # they must have selected that scale

                colorscale_selected = trace.meta[select_key]
                if auto or color_range==[None]:
                    trace.meta['color_range'] = (None)
                    if calc_key in trace.meta: del trace.meta[calc_key]
                    if select_key in trace.meta: del trace.meta[select_key]
                    trace.marker.colorscale = colorscale_selected
                    return

                data_min = np.nanmin(trace.marker.color)
                data_max = np.nanmax(trace.marker.color)
                if lock:
                    color_range = [data_min, data_max]
                trace.meta['color_range'] = color_range
                if color_range[0] is None: color_range[0] = data_min
                if color_range[1] is None: color_range[1] = data_max

                normalized_pairs = self._xp_renderer.project_color_scale(data_max, data_min, color_range, colorscale_selected)

                trace.marker.colorscale = tuple(normalized_pairs)
                trace.meta[calc_key] = tuple(normalized_pairs)
                # should ctime be the min max for the whole thing

class CrossPlot():
    marker_symbols = ["circle", "diamond", "square", "cross", "x", "pentagon", "start", "hexagram", "starsquare"]
    default_marker = {
            "opacity": .8,
            "size": 5
            }
    def _get_marker_no_color(self):
        marker = self.default_marker.copy()
        marker["symbol"] = self.marker_symbols[len(self.marker_symbols) % self._marker_symbol_index]
        self._marker_symbol_index += 1
        return marker

    def _get_marker_with_color(self, array, title=None, colorscale="Viridis_r", container_width=None, size=None):
        if not size: size = self.size
        marker = self._get_marker_no_color()
        marker["color"] = array
        marker["showscale"] = True
        marker["colorscale"] = colorscale
        if container_width is not None:
            marker["colorbar"] = dict( # again, problem here, this should be in layout, but its in trace, and that sucks
                    title=title,
                    orientation='h',
                    thickness=20,
                    thicknessmode='pixels',
                    x=(self.size/(2.00)-45)/container_width,
                    xref='paper',
                    y=10/size, # 10% margin, does this really work for all sizes?
                    yref='paper',
                    len=size*.9,
                    lenmode='pixels')
        else:
            marker["colorbar"] = dict(
                    title=title,
                    orientation='h',
                    thickness=20,
                    thicknessmode='pixels')
        marker["opacity"] = self.fade
        return marker

    def _resolve_selector_to_data(self, selector):
        POZO_OBJS = (pozo.Graph, pozo.Track, pozo.Axis)
        if isinstance(selector, POZO_OBJS):
           data = selector.get_data()
           if len(data) == 0:
               raise ValueError(f"{selector} has no pozo.Trace object")
           return data[0] # we process it in the following
        elif isinstance(selector, pozo.Trace):
            if not is_array(selector.get_data()): raise TypeError(f"{selector}'s data seems weird: {selector.get_data()}")
            if not is_array(selector.get_depth()):raise TypeError(f"{selector}'s depth seems weird: {selector.get_depth()}")
            return selector
        raise TypeError(f"{selector} does not appear to be a pozo object, it's a {type(selector)}")

    def reinit(self, x = None, y = None, colors=[None], **kwargs):
        self.size                = kwargs.pop("size", self.size)
        self.depth_range         = kwargs.pop("depth_range", self.depth_range)
        self.xrange              = kwargs.pop("xrange", self.xrange)
        self.yrange              = kwargs.pop("yrange", self.yrange)
        if not is_array(colors): colors = [colors]
        self.colors = colors
        self.y              = y if y is not None else self.y
        self.x              = x if x is not None else self.x

    def __init__(self, x=None, y=None, colors=[None], **kwargs):
        # rendering defaults
        self.size                = kwargs.pop("size", 500)
        self.depth_range         = kwargs.pop("depth_range", [None])
        self.xrange              = kwargs.pop("xrange", None)
        self.yrange              = kwargs.pop("yrange", None)
        self.fade = 1
        if not is_array(colors): colors = [colors]
        self.colors = colors
        self.y = y
        self.x = x
        self._colors_by_id = {}
        self._figures_by_id = weakref.WeakValueDictionary() # Do figures contain their colors? so we can clear up _colors_by_id

    def add_figure(self, fig):
        self._figures_by_id[id(fig)] = fig

    def render(self, **kwargs):
        static = kwargs.pop("static", False)
        layout = self.create_layout(**kwargs) # container_width, size
        traces = self.create_traces(**kwargs) # container_width, depth_range, size, static

        if static:
            self.last_fig = go.Figure(data=traces, layout=layout)
            return self.last_fig
        fig = xpFigureWidget(data=traces, layout=layout, renderer=self)
        self.add_figure(fig)
        self.last_fig = fig

        return fig

    def create_layout(self, container_width=None, size=None, xaxis="xaxis1", yaxis="yaxis1"):
        if not size: size = self.size
        margin = (120) / size if container_width is not None else 0
        return {
            "width"       : size,
            "height"      : size,
            xaxis       : dict(
                            title = self.x.get_name(),
                            range = self.xrange,
                            linecolor = "#888",
                            linewidth = 1,
            ),
            yaxis       : dict(
                            title = self.y.get_name(),
                            range = self.yrange,
                            domain = (margin, 1),
                            linecolor = "#888",
                            linewidth = 1,
            ),
            "showlegend"  : True
        }

    def create_traces(self, depth_range=None, container_width=None, size=None, static=False, xaxis="xaxis1", yaxis="yaxis1", color_lock={}, by_index=False):
        if not size: size = self.size
        if not depth_range: depth_range = self.depth_range
        x_data = self.x.get_data()[slice(*depth_range)] if by_index else self.x.get_data(slice_by_depth=depth_range)
        y_data = self.y.get_data()[slice(*depth_range)] if by_index else self.y.get_data(slice_by_depth=depth_range)

        self._marker_symbol_index = 1

        # Doing some stats
        missing = (np.isnan(x_data) + np.isnan(y_data)).sum() # noqa TODO
        # TODO improve with more statistics, how many values are there, etc
        title = "Number of unplottable values: {missing} ({(100 * missing/len(x_data)):.1f}%)" # noqa TODO
        # title will be added but also updated whenever written so maybe this should be a function

        self._base_trace = dict(
            x = x_data,
            y = y_data,
            mode='markers',
            xaxis=toTarget(xaxis),
            yaxis=toTarget(yaxis),
        )

        # Make Traces
        trace_definitions = []
        plotly_traces     = []
        for color in self.colors:
            trace_definitions.append(self.create_trace(color, container_width=container_width, depth_range=depth_range, size=size, static=static, by_index=by_index))
        if trace_definitions and 'visible' in trace_definitions[0]: del trace_definitions[0]['visible']

        for trace in trace_definitions:
            plotly_traces.append(go.Scattergl(trace))
            if trace['name'] in color_lock:
                # originally we did this post-process modification  because we didn't have plotly.colors.get_colorscale(str)
                # we could now do it in create trace
                color_range = color_lock[trace['name']]
                cs_sel = plotly_traces[-1]['marker']['colorscale'] # this is the selected
                data_min = np.nanmin(plotly_traces[-1]['marker']['color'])
                data_max = np.nanmax(plotly_traces[-1]['marker']['color'])
                cs_calc = self.project_color_scale(data_max, data_min, color_range, cs_sel)
                plotly_traces[-1]['marker']['colorscale'] = cs_calc
                plotly_traces[-1].meta['color_range'] = tuple(color_range)
                plotly_traces[-1].meta['colorscale_calculated'] = tuple(cs_calc)
                plotly_traces[-1].meta['colorscale_selected'] = tuple(cs_sel)
                # this should lock it, otherwise, it goes back to auto
        return plotly_traces


    def create_trace(self, color, container_width=None, depth_range=None, size=None, static=False, by_index=False):
        if not size: size = self.size
        trace = self._base_trace.copy()
        trace['meta']={'filter':'depth'}
        if color is not None:
            if isinstance(color, str) and color.lower() == "depth":
                trace['meta']['color_data'] = 'depth'
                color_array = self.x.get_depth()[slice(*depth_range)] if by_index else self.x.get_depth(slice_by_depth=depth_range)
            else:
                color_data = self._resolve_selector_to_data(color)
                color_array = color_data.get_data()[slice(*depth_range)] if by_index else color_data.get_data(slice_by_depth=depth_range)
                if not static: # no need to store if image will never update
                    trace['meta']['color_data'] = id(color)
                    self._colors_by_id[id(color)] = color

            name = color_data.get_name() if color != "depth" else "depth"
            trace['name'] = name
            trace['marker'] = self._get_marker_with_color(color_array, name, container_width=container_width, size=size)
            trace['hovertemplate'] = '%{x}, %{y}, %{marker.color}'
            trace['visible'] = 'legendonly'
        else:
            trace['marker'] = self._get_marker_no_color()
            trace['name'] = "x"
        return trace

    def project_color_scale(self, data_max, data_min, color_range, colorscale_selected):
        distance = data_max - data_min
        normalized_color_range = np.round((color_range - data_min) / distance, 2)
        norm_distance = normalized_color_range[1] - normalized_color_range[0]

        normalized_pairs = [] # lets create normalized pairs
        for pair in colorscale_selected:
            normalized_pairs.append(((pair[0]*norm_distance)+normalized_color_range[0], pair[1]))
        if normalized_pairs[0][0] > 1: normalized_pairs = [(0, normalized_pairs[0][1]), (1,normalized_pairs[0][1]) ]
        elif normalized_pairs[-1][0] < 0: normalized_pairs = [(0, normalized_pairs[-1][1]), (1,normalized_pairs[-1][1]) ]
        else:
            # can bottom be extended
            if normalized_pairs[0][0] > 0: normalized_pairs.insert(0, (0, normalized_pairs[0][1]))
            elif normalized_pairs[0][0] < 0:
                lb = normalized_pairs[0]
                ub = None
                old_pairs = normalized_pairs.copy()
                for pair in old_pairs:
                    if pair[0] > 0:
                        if ub is None:
                            ub = pair
                            #print(lb)
                            new_color = plotly.colors.label_rgb(
                                plotly.colors.find_intermediate_color(
                                    plotly.colors.hex_to_rgb(lb[1]) if lb[1][0] != 'r' else plotly.colors.unlabel_rgb(lb[1]),
                                    plotly.colors.hex_to_rgb(ub[1]) if ub[1][0] != 'r' else plotly.colors.unlabel_rgb(ub[1]),
                                    (-lb[0])/(ub[0]-lb[0]) # distance from 0
                                    ))
                            #print(ub)
                            normalized_pairs = [(0, new_color)]
                        normalized_pairs.append(pair)
                    else:
                        lb = pair
            if normalized_pairs[-1][0] < 1: normalized_pairs.append((1, normalized_pairs[-1][1]))
            elif normalized_pairs[-1][0] > 1:
                ub = normalized_pairs[-1]
                lb = None
                old_pairs = list(reversed(normalized_pairs.copy()))
                for pair in old_pairs:
                    if pair[0] < 1: # we're here
                        if lb is None:
                            lb = pair
                            #print(lb)
                            #print(ub)
                            new_color = plotly.colors.label_rgb(
                                plotly.colors.find_intermediate_color(
                                    plotly.colors.hex_to_rgb(lb[1]) if lb[1][0] != 'r' else plotly.colors.unlabel_rgb(lb[1]),
                                    plotly.colors.hex_to_rgb(ub[1]) if ub[1][0] != 'r' else plotly.colors.unlabel_rgb(ub[1]),
                                    (1-lb[0])/(ub[0]-lb[0]) # distance from 0
                                    ))
                            normalized_pairs = [(1, new_color)]
                        normalized_pairs.append(pair)
                    else:
                        ub = pair
                normalized_pairs = list(reversed(normalized_pairs))
        return normalized_pairs

    @property
    def colors(self):
        return self.__colors

    @colors.setter
    def colors(self, colors):
        self.__colors = []
        for color in colors:
            if color is None:
                self.__colors.append(None)
            elif isinstance(color, str) and color.lower() == "depth":
                self.__colors.append("depth")
            else:
                self.__colors.append(self._resolve_selector_to_data(color))

    @property
    def x(self):
        return self.__x
    @x.setter
    def x(self, x):
        self.__x = self._resolve_selector_to_data(x) if x is not None else None

    @property
    def y(self):
        return self.__y
    @y.setter
    def y(self, y):
        self.__y = self._resolve_selector_to_data(y) if y is not None else None


def init_write_image(counter):
    global write_image_counter
    write_image_counter = counter

def write_image(gp):
    gp['graph'].write_image(gp['path'], engine="kaleido")
    with write_image_counter.get_lock():
        write_image_counter.value += 1

def make_xp_depth_video(folder_name, graph, start, window, end, xp=True, output="all.mp4", fps=30):
    try:
        write_image_counter
        raise Exception("Please only run make_frames once at a time, or restart the kernel")
    except NameError:
        pass
    if output:
        try:
            import imageio
        except ImportError:
            raise ImportError("Please install imageio. It must be installed like: pip install imageio[ffmpeg] or pip install imageio[pyav]")
    os.makedirs(folder_name, exist_ok=True)
    if xp is True: xp=graph.xp
    elif not isinstance(xp, CrossPlot): raise ValueError("Crossplot renderer not valid. xp=True (default) or another renderer)")
    trace = xp.x
    depth = trace.get_depth()
    start_index = trace.find_nearest(start)[0]
    window_index = trace.find_nearest(start+window)[0] - start_index
    end_index = trace.find_nearest(end)[0]
    frame_count = range(start_index, end_index-window_index, 1)
    gp = []
    render_counter = IntProgress(min=0, max=len(frame_count)-1, description="Rendering:")
    writer_counter = IntProgress(min=0, max=len(frame_count)-1, description="Writing:")
    display(render_counter)
    display(writer_counter)
    tail_size = math.floor(window_index*.2)
    tail = np.linspace(0,1,tail_size)
    fade = tail.tolist() + [1]*(window_index-tail_size)
    for i, cursor in enumerate(frame_count):
        render_counter.value += 1
        graph.depth_notes['Depth Highlight-xxx'] = dict(range = (depth[cursor], depth[cursor+window_index]), show_text=False)
        graph.xp.fade = fade
        gp.append(
                dict(
                    graph=graph.render(
                        height=800,
                        xp=True,
                        depth_position=1,
                        color_lock={'depth': [start, end]},
                        depth=[start, end],
                        xp_depth_by_index=[cursor, cursor+window_index],
                        static=True,
                        ),
                    path=folder_name+"/"+str(i)+".png",
                    )
                )
    del graph.depth_notes['Depth Highlight-xxx']
    write_image_counter = multiprocessing.Value('I',0)
    with multiprocessing.Pool(initializer=init_write_image, initargs=(write_image_counter,)) as pool:
        p = pool.map_async(write_image, gp)
        while True:
            p.wait(.5)
            try:
                p.successful()
                break
            except ValueError:
                writer_counter.value = write_image_counter.value
    writer_counter.value = len(frame_count)-1
    del write_image_counter
    if not output: return
    ims = [imageio.imread(p['path']) for p in gp]
    imageio.mimwrite(folder_name+"/" + output, ims, fps=fps)



import colour

import pozo.themes.mnemonic_tables as tables

class Theme(): # Meant to be inherited
    def __init__(self, *args, **kwargs):
        self._context = {}
        super().__init__(*args, **kwargs)

    def set_context(self, data):
        if data is None: data = {}
        if not isinstance(data, dict):
            raise TypeError("Cannot assing context which is not a dict (or None)")
        self._context = data

    def get_context(self):
        return self._context

    def resolve(self, key, contexts):
            raise NotImplementedError("Every Theme inheriter must supply a resolve(self, key, context)")

class ThemeDict(Theme, dict): # Basically a dict
    def resolve(self, key, contexts=None):
        if key not in self:
            return None
        return self[key]

class DynamicTheme(Theme): # Meant to be inherited
    def resolve(self, key, contexts):
        raise NotImplementedError("Dynamic themes must implement get_value(self, key, args)")

# A themeable is something that stores a theme. It accepts either a theme or a dict. It sets its metadata on the theme
# When
class Themeable(): # Meant to be inherited by objects
    def __init__(self, *args, **kwargs):
        self.set_theme(kwargs.pop("theme", {}))
        super().__init__(*args, **kwargs)

    def set_theme(self, theme):
        if isinstance(theme, str):
            if theme not in themes:
                raise ValueError(f"{theme} is not a built-in theme, options are: {', '.join(list(themes.keys()))}")
            theme = themes[theme]
        if isinstance(theme, dict) and not isinstance(theme, ThemeDict):
            self._theme = ThemeDict(theme)
        elif isinstance(theme, Theme):
            self._theme = theme
        elif theme is None:
            self._theme = ThemeDict({})
        else:
            raise TypeError(f"Theme was not a dict, ThemeDict, or DynamicTheme, was {type(theme)}")

    def _get_theme(self, context={}):
        self._theme.set_context(context)
        return self._theme

    # You may override this
    def get_theme(self):
        return self._get_theme()

# Above are inheritables
from pozo.themes.theme_tools import * # should it be relative?
# Below is implementations
themes = {'cangrejo': MnemonicDictionary(tables.cangrejo)}

# ThemeList also has a theme, which is considered an override!
class ThemeStack(Themeable):
    def __init__(self, *args, **kwargs):
        self._contexts_vertical = [] # parallel con _list
        self._list = [] # the themes
        self._shadow_objects = {} # if themes have things that are initialized when they're added
        for arg in args:
            self.append(arg)
        super().__init__(*[], **kwargs)

    def append(self, theme):
        if isinstance(theme, Theme):
            self._list.append(theme)
            self._contexts_vertical.append(theme.get_context()) # but why not just traverse this at get time
        else:
            raise TypeError(f"Found a theme which isn't any type of valid theme: {type(theme)} {theme}")

    def pop(self, index = -1):
        self._contexts_vertical.pop(index)
        return self._list.pop(index)

    def __getitem__(self, key):
        if not isinstance(key, str):
            raise TyperError("Key must be string")
        # print(f'\ngetitem() Trying to get theme: {key}')

        contexts_horizontal = []
        eventual_value = self.get_theme() # Self-theme, override
        eventual_value = self._check_shortcut(eventual_value, key, contexts_horizontal) # something that translates to themelike
        # print(f'Override: {eventual_value}')
        while True: # do-while loop using break
            contexts_horizontal.append(eventual_value.get_context())
            eventual_value = eventual_value.resolve(key, self._contexts_vertical)
            # print(f'Override resolved to: {eventual_value}')
            eventual_value = self._check_shortcut(eventual_value, key, contexts_horizontal) # something that translates to themelike
            # print(f'Above shortcutted to: {eventual_value}')
            if not isinstance(eventual_value, Theme):
                break
        if eventual_value is None:
            contexts_horizontal.clear()
            for eventual_value in reversed(self._list):
                # print(f'Value from list of length {len(self._list)}: {eventual_value}')
                while True:# do-while loop using break
                    contexts_horizontal.append(eventual_value.get_context())
                    eventual_value = eventual_value.resolve(key, self._contexts_vertical)
                    # print(f'Override resolved to: {eventual_value}')
                    eventual_value = self._check_shortcut(eventual_value, key, contexts_horizontal)
                    # print(f'Above shortcutted to: {eventual_value}')
                    if not isinstance(eventual_value, Theme):
                        break
                if eventual_value is not None:
                   break
        # print(f'Final result: {eventual_value}')
        return self._process_output(eventual_value, key)

    # Take not-themes and return themes. Can't take a regular dict, which is sorta problematic.
    def _check_shortcut(self, value, key, contexts):
        context = contexts[-1] if len(contexts) else []
        if value is None:
            return None
        may_be_key = (id(value), id(context))
        if may_be_key in self._shadow_objects:
            return self._shadow_objects[may_be_key]
        elif key == "color" and isinstance(value, list):
            value = ColorWheel(value, context=context)
        elif key == "place_holder" and isinstance(value, "placeholder"): # never happen, put other shortcuts here
            value = "place_holder"
        else:
            return value # just send back what we got, do nothing
        self._shadow_objects[may_be_key] = value
        return value

    def _process_output(self, value, key):
        if value is None:
            return None # default should supply all values. Grep "How did this happen" to find other error. TODO
        if key == "color":
            if not isinstance(value, colour.Color):
                return colour.Color(value).hex_l
            else:
                return value.hex_l
        return value




import pozo.themes as pzt
import pozo.units as pzu


default_color_list = ['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A', '#19D3F3', '#FF6692', '#B6E880', '#FF97FF', '#FECB52']
default_theme = pzt.ThemeDict({ "color": default_color_list, "track_width": 200 })
#
class MnemonicDictionary(pzt.DynamicTheme):
    def __init__(self, mnemonic_table, registry=pzu.registry):
        if not isinstance(mnemonic_table, dict): raise ValueError("menmonic_table must be dictionary")
        for key, value in mnemonic_table.items():
            if "range_unit" in value:
                pzu.registry.parse_units(value["range_unit"]) # just checking to see if its working
        self._lut = mnemonic_table # renderer will have to do conversons

    def resolve(self, key, contexts):
        mnemonic = None
        for context in contexts:
            if "mnemonic" in context:
                mnemonic = context["mnemonic"]
            elif "mnemonics" in context and len(context["mnemonics"]) >= 1:
                    mnemonic = context["mnemonics"][0]
        if mnemonic is None or mnemonic not in self._lut:
            if '-' in self._lut and key in self._lut['-']: return self._lut['-'][key]
            return None
        if key in self._lut[mnemonic]:
            return self._lut[mnemonic][key]
        if '-' in self._lut and key in self._lut['-']: return self._lut['-'][key]
        return None

    def get_dictionary(self):
        return self._lut

    def set_mnemonic(self, mnemonic, dictionary):
        self._lut[mnemonic] = dictionary

    def get_mnemonic(self, mnemonic):
        if mnemonic not in self._lut: return None
        return self._lut[mnemonic]

    def set_value(self, mnemonic, key, value):
        if mnemonic not in self._lut: self._lut[mnemonic] = {}
        self._lut[mnemonic][key] = value

    def set_fallback(self, key, value):
        self.set_value('-', key, value)

class ColorWheel(pzt.DynamicTheme):
    def __init__(self, color_list=default_color_list, per=None, context=None, each=False):
        self._color_list=color_list
        self._each = each
        if per is None and context is not None and "type" in context:
            if context["type"] == "axis":
                self._per = "trace"
            else:
                self._per = "axis"
        elif per is None:
            self._per= "axis"
        else:
            self._per = per
        self._current_color = color_list[0]
        self._current_context = None
        self._i = 0
        super().__init__()

    def get_current_color(self):
        return self._current_color

    # ColorWheel is really married to graphs. Think about this during documentation.
    def resolve(self, key, contexts):
        if not self._each:
            for context in contexts:
                if "type" in context and (context["type"] == self._per or context["type"] in self._per):
                    if self._current_context is None:
                        self._current_context = id(context)
                    elif id(context) != self._current_context:
                        self._current_context = id(context)
                        color = self._color_list[self._i % len(self._color_list)]
                        self._current_color = color
                    else: # same context, do nothing
                        continue
                    self._i += 1
                    break
        else:
            color = self._color_list[self._i % len(self._color_list)]
            self._current_color = color
            self._i += 1
        return self._current_color




import pytest
import pozo.traces

index = [1, 2, 3, 4]
values = [0, 1, 1, 0]
index2 = [5, 6, 7, 8]
values2 = [2, 3, 4, 5]

d1 = pozo.traces.Trace(values, depth=index, mnemonic="md1")
d2 = pozo.traces.Trace(values, depth=index, name="d2")
d3 = pozo.traces.Trace(values, depth=index, name="d3", mnemonic="md3")
d4 = pozo.traces.Trace(values, depth=index, mnemonic="md4")
d5 = pozo.traces.Trace(values, depth=index, mnemonic="md5")
d6 = pozo.traces.Trace(values, depth=index, mnemonic="md6")
d6 = pozo.traces.Trace(values, depth=index, mnemonic="md7")
d7 = pozo.traces.Trace(values, depth=index, mnemonic="md8")

def test_user_simulation():
    x = pozo.traces.Trace(values, depth=index, mnemonic="DHO")
    assert isinstance(x, pozo.traces.Trace)
    assert x.get_name() == "DHO"
    x.set_name("NAME")
    assert x.get_name() == "NAME"
    assert x.get_mnemonic() == "DHO"
    x.set_mnemonic("TEST")
    assert x.get_mnemonic() == "TEST"
    assert x.get_data() == values
    assert x.get_depth() == index
    x.set_data(values2)
    x.set_depth(index2)
    assert x.get_data() == values2
    assert x.get_depth() == index2
    # Error
    with pytest.raises(ValueError):
        y = pozo.traces.Trace(values, depth=index)
    with pytest.raises(ValueError):
        x.set_data([1])
    with pytest.raises(ValueError):
        x.set_depth([1])
    with pytest.raises(ValueError):
        x.set_data([1,2], depth=[1])

def test_user_simulation2():
    ## Init
    first = pozo.traces.Trace([2], depth=[1], name="test", mnemonic="test2")
    assert isinstance(first, pozo.traces.Trace)
    assert first.get_data() == [2]
    assert first.get_depth() == [1]
    assert first.get_name() == "test"
    assert first.get_mnemonic() == "test2"
    second = pozo.traces.Trace([2], depth=[1], name="test")
    assert isinstance(second, pozo.traces.Trace)
    assert second.get_data() == [2]
    assert second.get_depth() == [1]
    assert second.get_name() == "test"
    assert second.get_mnemonic() is None
    second.set_mnemonic("test2")
    assert second.get_mnemonic() == "test2"
    second.set_name("test3")
    assert second.get_name() == "test3"
    third = pozo.traces.Trace([2], depth=[1], mnemonic="test2")
    assert isinstance(third, pozo.traces.Trace)
    assert third.get_data() == [2]
    assert third.get_depth() == [1]
    assert third.get_name() == "test2"
    assert third.get_mnemonic() == "test2"
    with pytest.raises(Exception):
        pozo.traces.Trace([1], depth=[1,2], name="what")
    with pytest.raises(Exception):
        pozo.traces.Trace([1, 2], depth=[1,2])




import warnings
import ood
import pozo
import pozo.renderers as pzr
import pozo.themes as pzt
import traceback


class Axis(ood.Item, pzt.Themeable):
    _type = "axis"
    _child_type = "trace"

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        for ar in args:
            self.add_traces(ar)

    def _check_types(self, *traces):
        accepted_types = (pozo.Trace)
        raw_return = []
        for trace in traces:
            if isinstance(trace, list):
                raw_return.extend(self._check_types(*trace))
            elif not isinstance(trace, accepted_types):
                raise TypeError(f"Axis.add_traces() only accepts pozo.Trace, or a single list of pozo.Trace, not {type(trace)}")
            else:
                raw_return.append(trace)
        return raw_return

    # add_items
    def add_traces(self, *traces, **kwargs):
        good_traces = self._check_types(*traces)
        super().add_items(*good_traces, **kwargs)
        return good_traces

    # get_items
    def get_traces(self, *selectors, **kwargs):
        return super().get_items(*selectors, **kwargs)

    # get_item
    def get_trace(self, selector, **kwargs):
        return super().get_item(selector, **kwargs)

    # pop items
    def pop_traces(self,  *selectors, **kwargs):
        return super().pop_items(*selectors, **kwargs)

    # what about whitelabelling all the other stuff
    def has_trace(self, selector):
        return super().has_item(selector)

    def reorder_all_traces(self, order):
        super().reorder_all_items(order)

    def move_traces(self, *selectors, **kwargs):
        super().move_items(*selectors, **kwargs)

    def get_named_tree(self):
        result = []
        for el in self.get_traces():
            result.append(el.get_named_tree())
        return { "axis" : { self.name: result } }

    def get_theme(self):
        mnemonics = []
        for d in self.get_traces():
            mnemonics.append(d.get_mnemonic())
        context = { "type":"axis",
                   "name": self._name,
                   "mnemonics": mnemonics,
                   }
        return self._get_theme(context=context)



import warnings
import pint
import numpy as np
import ood
import pozo.themes as pzt
import pozo.units as pzu


class Trace(ood.Observed, pzt.Themeable):
    def __len__(self):
        return len(self.get_data())

    def __init__(self, data, **kwargs):
        unit = kwargs.pop("unit", None)
        depth = kwargs.pop("depth", None)
        depth_unit = kwargs.pop("depth_unit", None)
        self._unit = None
        self._depth_unit = None
        if depth is None:
            raise ValueError("`depth` cannot be None, you must supply a depth")
        if len(depth) != len(data):
            raise ValueError("Depth and values have different length")

        self._mnemonic = kwargs.pop("mnemonic", None)
        if "name" not in kwargs:
            if not self._mnemonic:
                raise ValueError(
                    "You must supply 'name'. Or 'mnemonic' will be used as 'name' if 'name' absent..."
                )
            kwargs["name"] = self._mnemonic

        self.set_data(data, unit=unit, depth=depth, depth_unit=depth_unit)
        super().__init__(**kwargs)  # od.ChildObserved sets name

    def set_data(self, data, unit=None, depth=None, depth_unit=None):
        unit = self._check_unit(unit)
        depth_unit = self._check_unit(depth_unit)
        depth_len = len(depth) if depth is not None else len(self._depth)
        if depth_len != len(data):
            raise ValueError("Depth and values have different length.")

        self._data = data
        if unit is not None:
            self.set_unit(unit)
        elif isinstance(data, pint.Quantity):
            self.set_unit(data.units)
        # otherwise keeps old unit

        if depth is not None:
            self.set_depth(depth, depth_unit=depth_unit)  # this will set the depth unit

    def _find_nearest(self, value):
        # TODO: align units depth_range, assume same units, but check if pint
        array = np.asarray(self.get_depth())
        idx = np.nanargmin((np.abs(array - value)))
        # if array[idx] != value:
        #    raise ValueError(f"Tried to find value {value} in {self.get_name()} but the closest value was {array[idx]}")
        return idx

    def find_nearest(self, value):
        # TODO: align units depth_range, assume same units, but check if pint
        array = np.asarray(self.get_depth())
        idx = np.nanargmin((np.abs(array - value)))
        # if array[idx] != value:
        #    raise ValueError(f"Tried to find value {value} in {self.get_name()} but the closest value was {array[idx]}")
        return idx, array[idx]

    def get_data(self, slice_by_depth=None):
        if not slice_by_depth or slice_by_depth == [None]:
            return self._data
        return self._data[
            slice(
                *[
                    self._find_nearest(val) if val is not None else val
                    for val in slice_by_depth
                ]
            )
        ]

    def set_depth(self, depth, depth_unit=None):
        depth_unit = self._check_unit(depth_unit)
        if len(self._data) != len(depth):
            raise ValueError("Depth and values have different length.")

        self._depth = depth
        if depth_unit is not None:
            self.set_depth_unit(depth_unit)
        elif isinstance(depth, pint.Quantity):
            self.set_depth_unit(depth.units)
        # else, keep old units

    def get_depth(self, slice_by_depth=None):
        if not slice_by_depth or slice_by_depth == [None]:
            return self._depth
        return self._depth[
            slice(
                *[
                    self._find_nearest(val) if val is not None else val
                    for val in slice_by_depth
                ]
            )
        ]

    def _check_unit(
        self, unit
    ):  # Call this early, in set_data, in set_depth, in init()
        if isinstance(unit, str):
            return pzu.registry.parse_units(unit)
        elif not isinstance(unit, pint.Unit) and unit is not None:
            raise pint.UndefinedUnitError(str(type(unit))) from TypeError(
                f"Unrecognized unit type: {unit}"
            )
        return unit

    def set_unit(self, unit):
        if isinstance(unit, str):
            unit = pzu.registry.parse_units(unit)
        elif isinstance(unit, pint.Unit) or unit is None:
            pass
        else:
            raise pint.UndefinedUnitError(str(type(unit))) from TypeError(
                f"Unrecognized unit type: {unit}"
            )
        self._unit = unit

    def get_unit(self):
        return self._unit

    def set_depth_unit(self, unit):
        if isinstance(unit, str):
            unit = pzu.registry.parse_units(unit)
        elif isinstance(unit, pint.Unit) or unit is None:
            pass
        else:
            raise pint.UndefinedUnitError(str(type(unit))) from TypeError(
                f"Unrecognized unit type: {unit}"
            )
        self._depth_unit = unit

    def get_depth_unit(self):
        return self._depth_unit

    def convert_depth_unit(self, unit):
        unit = self._check_unit(unit)
        if (
            unit == self.get_depth_unit()
            or unit is None
            or self.get_depth_unit() is None
        ):
            return
        if isinstance(self.get_depth(), pint.Quantity):
            self.set_depth(self.get_depth().to(unit))  # This would be another version
        else:
            self.set_depth(
                pzu.Q(self.get_depth(), self.get_depth_unit()).to(unit).magnitude
            )  # This would be another version
        self.set_depth_unit(unit)

    # not tested
    def convert_unit(self, unit):
        unit = self._check_unit(unit)
        if unit == self.get_unit() or unit is None or self.get_unit() is None:
            return
        if isinstance(self.get_data(), pint.Quantity):
            self.set_data(self.get_data().to(unit))  # This would be another version
        else:
            self.set_data(
                pzu.Q(self.get_data(), self.get_unit()).to(unit).magnitude
            )  # This would be another version
        self.set_unit(unit)

    def set_mnemonic(self, mnemonic):
        self._mnemonic = mnemonic

    def get_mnemonic(self):
        return self._mnemonic

    def get_named_tree(self):
        return {
            "trace": {
                "name": self._name,
                "mnemonic": self._mnemonic,
                "length": len(self._values),
            }
        }

    def get_theme(self):
        context = {
            "type": "trace",
            "name": self._name,
            "mnemonic": self._mnemonic,
        }
        return self._get_theme(context=context)


class Data(Trace):
    def __init__(self, data, **kwargs):
        warnings.warn("pozo.Data is deprecated, use pozo.Trace", DeprecationWarning)
        super().__init__(self, data, **kwargs)



from .units import *

registry = LasRegistry()
Q = registry.Quantity # Overriding the registry and all this is a little weird

registry.define('gamma_API_unit = [Gamma_Ray_Tool_Response]  = gAPI')
registry.define("porosity_unit = percent = pu")
registry.define("of_1 = 100 * percent = fraction")
registry.define("legacy_api_porosity_unit = [Legacy_API_Porosity_Unit] = puAPI")

PU = [
        LasMapEntry((-50, 80), "pu", "decide by range - MEDIUM"),
        LasMapEntry((-.5, .8), "of_1", "decide by range - MEDIUM"),
        LasMapEntry((), "puAPI", "catch all, legacy unit - LOW"),
    ]
percent_general = [
        LasMapEntry((-100, 100), "percent", "decide by range, verify - LOW"),
        LasMapEntry((-1, 1), "of_1", "decide by range, verify - LOW"),
        LasMapEntry((), "ppm", "catch all, ppm - LOW" ),
    ]

registry.add_las_map('DEPT', 'M'   , 'meter'                   , "HIGH")
registry.add_las_map('GR'  , 'GAPI', 'gAPI'                    , "HIGH")
registry.add_las_map('CGR' , 'GAPI', 'gAPI'                    , "HIGH")
registry.add_las_map('CALI', 'CM'  , 'centimeter'              , "HIGH")
registry.add_las_map('DRHO', 'G/C3', 'gram / centimeter ** 3'  , "HIGH")
registry.add_las_map('RHOB', 'G/C3', 'gram / centimeter ** 3'  , "HIGH")
registry.add_las_map('DT'  , 'US/F', 'microsecond / foot'      , "HIGH")
registry.add_las_map('ILD' , 'OHMM', 'ohm * meter'             , "HIGH")
registry.add_las_map('LLS' , 'OHMM', 'ohm * meter'             , "HIGH")
registry.add_las_map('ILM' , 'OHMM', 'ohm * meter'             , "HIGH")
registry.add_las_map('LLD' , 'OHMM', 'ohm * meter'             , "HIGH")
registry.add_las_map('SFL' , 'OHMM', 'ohm * meter'             , "HIGH")
registry.add_las_map('SGR' , 'GAPI', 'gAPI'                    , "HIGH")
registry.add_las_map('SP'  , 'MV'  , 'millivolt'               , "HIGH")
registry.add_las_map('MSFL', 'OHMM', 'ohm * meter'             , "HIGH")
registry.add_las_map('NPHI', 'PU'  , PU)
registry.add_las_map('NPHI', 'PU%' , 'pu'                      , "HIGH")
registry.add_las_map('PEF' , ''    , percent_general)
registry.add_las_map('POTA', ''    , percent_general)
registry.add_las_map('THOR', ''    , percent_general)
registry.add_las_map('URAN', ''    , percent_general)

# TODO: Beginning of a database, has to go online
# mnemonics, synonyms (resolve- matching in theme,
# matching in recipe, matching for unit determination)
# relationships between mnemonics? tolerances? tools?
# annotating reg files w/ corrections that the file doesn't support
# versioning, searching
# also want it to point to posts
import re
desc_wo_num = re.compile(r'^(?:\s*\d+\s+)?(.*)$')
red_low = re.compile(r'<td>(.+)?LOW(.+)?</td>')
orange_medium = re.compile(r'<td>(.+)?MEDIUM(.+)?</td>')

def check_las(las, registry=registry):
    def n0(s): # If None, convert to ""
        return "" if s is None else str(s)
    d = chr(0X1e) # delimiter
    result = [f"mnemonic{d}las unit{d}pozo mapping{d}confidence{d}parsed{d}description{d}min{d}med{d}max{d}#NaN"]
    for curve in las.curves:
        resolved = None
        pozo_match = None
        confidence = None
        try:
            resolved = registry.resolve_las_unit(curve.mnemonic, curve.unit, curve.data)
            if resolved is not None:
                pozo_match = resolved.unit
                confidence = resolved.confidence
            parsed = registry.parse_unit_from_context(curve.mnemonic, curve.unit, curve.data)
        except (pint.UndefinedUnitError, MissingRangeError) as e:
            confidence = " - " + str(e) + " - NONE"
        find_desc = desc_wo_num.findall(curve.descr)
        desc = find_desc[0] if len(find_desc) > 0 else curve.descr
        [min, med, max] = [str(x) for x in np.nanquantile(curve.data, [0, .5, 1])]
        result.append(d.join([n0(x) for x in [curve.mnemonic,
                                             curve.unit,
                                             pozo_match,
                                             confidence,
                                             parsed,
                                             desc, 
                                             min, med, max, np.count_nonzero(np.isnan(curve.data))]
                              ]))
    try:
        output = pd.read_csv(StringIO("\n".join(result)), delimiter=d, na_filter=False).to_html()
        for match in red_low.finditer(output):
            current_match = match.group()
            colored = "<td style=\"color:red\">" + current_match[4:]
            output = output.replace(current_match, colored)
        for match in orange_medium.finditer(output):
            current_match = match.group()
            colored = "<td style=\"color:#B95000\">" + current_match[4:]
            output = output.replace(current_match, colored)
        display(HTML(output))
    except Exception as e:
        display(str(e))
        display(HTML("<br>".join(result)))

# Just a shorcut to make the API nice
def parse_unit_from_curve(curve, registry=registry):
    return registry.parse_unit_from_context(curve.mnemonic, curve.unit, curve.data)



import pytest
import ood
import ood.selectors as s
import ood.exceptions as e
import pozo.traces, pozo.axes, pozo.tracks

index = [1, 2, 3, 4]
values = [0, 1, 1, 0]
index2 = [5, 6, 7, 8]
values2 = [2, 3, 4, 5]

e.NameConflictException.default_level = e.ErrorLevel.IGNORE
e.MultiParentException.default_level = e.ErrorLevel.IGNORE

d1 = pozo.traces.Trace(values, depth=index, mnemonic="md1")
d2 = pozo.traces.Trace(values, depth=index, name="d2")
d3 = pozo.traces.Trace(values2, depth=index, name="d3", mnemonic="md3")
d4 = pozo.traces.Trace(values2, depth=index, mnemonic="md4")
d5 = pozo.traces.Trace(values, depth=index2, mnemonic="md5")
d6 = pozo.traces.Trace(values, depth=index2, mnemonic="md6")
d7 = pozo.traces.Trace(values2, depth=index2, mnemonic="md7")
d8 = pozo.traces.Trace(values2, depth=index2, mnemonic="md8")
a1 = pozo.axes.Axis()
a2 = pozo.axes.Axis()
a3 = pozo.axes.Axis(d8)
a4 = pozo.axes.Axis(name="I have a name")
a5 = pozo.axes.Axis()
a6 = pozo.axes.Axis()
a7 = pozo.axes.Axis()
a8 = pozo.axes.Axis()


def test_user_simulation():

    wdata = pozo.tracks.Track(d1, d2, d3, d4, d5, d6, d7, d8, a1, a2, a3, a4)
    assert wdata.get_axes(a1) == [a1]
    assert wdata.get_axes("unnamed") == [a1, a2, a3]
    assert wdata.get_axes(s.Has_Children(d1)) == wdata.get_axes("md1")
    assert wdata.get_axes(s.Has_Children(d8)) == [wdata.get_axis("md8"), wdata.get_axis(a3)]
    a = pozo.tracks.Track()
    assert isinstance(a, pozo.tracks.Track)
    assert a.get_axes() == []
    with pytest.raises(ValueError):
        a.get_axis(None)
    assert a.pop_axes() == []
    b = pozo.tracks.Track(a1)
    assert len(b.get_axes()) ==1
    assert len(b) == 1
    assert b.get_axis(a1.get_name()) == a1
    assert b.get_axis(a1) == a1
    assert b.get_axis(a2) == None
    assert b.get_axis('astoasi') == None
    assert b.get_axis(5) == None
    c = pozo.tracks.Track(*[a1, a2])
    d = pozo.tracks.Track(*[a1, a2], a3)
    e = pozo.tracks.Track(*[a4, a5], a6)
    assert e._count_dictionary() == 3
    assert len(e.get_axes()) == 3
    assert e._count_dictionary() == len(e.get_axes())
    assert e.get_axes() == [a4, a5, a6]
    assert e.get_axis(a4) == a4
    assert e.get_axis(a1) == None
    assert e.get_axis(a2) == None
    assert e.get_axis('astoasi') == None
    assert e.get_axis(5) == None

    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_axes())
    with pytest.warns(UserWarning):
        e.add_axes(*[a5, a1]) # a1 new
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_axes())
    assert len(e.get_axes()) == 4
    with pytest.warns(UserWarning):
        e.add_axes(a1, a8, *[a2, a4]) #a2 a8 new
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_axes())
    assert len(e.get_axes()) == 6
    with pytest.warns(UserWarning):
        e.add_axes(*c.get_axes(), *d.get_axes(), *e.get_axes())
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_axes())
    assert len(e.get_axes()) == 7
    assert len(e.get_axes()) != 6
    e.add_axes()
    with pytest.raises(TypeError):
        e.add_axes(None)
        e.add_axes("whatever")
    assert e.get_axes(a6, a7) == [a6]
    assert e.get_axes(a6) == [a6]
    n = len(e.get_axes())
    assert e.pop_axes(a6) == [a6]
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_axes())
    assert n == len(e.get_axes()) + 1
    #assert e.pop_axes(a6) == [] # by default strict
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_axes())
    #assert e.pop_axes(a6, a6, a6) == [] # by default strict
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_axes())
    e.pop_axes(*e.get_axes())
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_axes())
    assert len(e.get_axes()) == 0
    with pytest.warns(UserWarning):
        e.add_axes(*[a1,a2],*[a3,a4],*[a1,a4])
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_axes())
    assert len(e.get_axes()) == 4
    assert e.get_axes(*[a1, a2, a3, a4]) == [a1, a2, a3, a4]
    assert e.get_axes(*[a1, a2],*[ a3, a4, a5]) == [a1, a2, a3, a4]
    assert e.get_axis(a1) == a1
    a1.set_name("test")
    assert e.get_axis("test") == a1



cangrejo  = {
    "CALI" : {
        "color" : "black",
        "range": [0, 16],
        "range_unit": "inch",
    },
    "CGR" : {
        "color" : "green",
        "range": [0, 150.00],
        "range_unit": "gAPI",
    },
    "LLS" : {
        "color" : "gray",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit": "ohm * meter",
    },
    "ILM" : {
        "color" : "blue",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit":  "ohm * meter",
    },
    "ILD" : {
        "color" : "black",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit":  "ohm * meter",
    },
    "LLD" : {
        "color" : "red",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit":  "ohm * meter",
    },
    "LLD" : {
        "color" : "red",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit":  "ohm * meter",
    },
    "LLD" : {
        "color" : "red",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit": "ohm * meter",
    },
    "NPHI" : {
        "color" : "#008080",
        "range": [0.45, -.15],
        "range_unit": "fraction",
    },
    "RHOB" : {
        "color" : "red",
        "range": [1.95, 2.95],
        "range_unit": "gram/centimeter**3",
    },
    "DT" : {
        "color" : "red",
        "range": [30, 200],
        "range_unit": "microsecond/foot",
    },
}



import warnings
import pozo
import pozo.units as pzu
import pozo.renderers as pzr
import pozo.themes as pzt
import ood
import ood.exceptions as ooderr

LAS_TYPE = "<class 'lasio.las.LASFile'>"  # TODO this isn't going to work reliably
WELLY_WELL_TYPE = "<class 'welly.well.Well'>"
WELLY_PROJECT_TYPE = "<class 'welly.project.Project'>"


# TODO we want to make sure that your depth has no NAN values (and warn user)
class Graph(ood.Observer, pzt.Themeable):
    _type = "graph"
    _child_type = "track"

    def __init__(self, *args, **kwargs):

        # Be cool if we could use include to specify things be on the same track TODO
        self._name = kwargs.pop("name", "unnamed")
        self.renderer = kwargs.pop("renderer", pzr.Plotly())
        self.xp = kwargs.pop(
            "xp", pzr.CrossPlot()
        )  # kinda don't like doing this, making it point to a class

        my_kwargs = {}  # Don't pass these to super, but still pass them down as kwargs
        my_kwargs["include"] = kwargs.pop("include", None)
        my_kwargs["exclude"] = kwargs.pop("exclude", None)
        my_kwargs["compare"] = kwargs.pop("compare", False)
        my_kwargs["yaxis"] = kwargs.pop("yaxis", None)
        my_kwargs["yaxis_name"] = kwargs.pop("yaxis_name", None)
        my_kwargs["unit_map"] = kwargs.pop("unit_map", None)
        old_kwargs = kwargs.copy()
        if not isinstance(self._name, str):
            raise TypeError("Name must be a string")
        try:
            super().__init__(**kwargs)
        except TypeError as te:
            raise TypeError(
                f"One of the arguments here isn't valid: {list(old_kwargs.keys())}."
            ) from te
        self.process_data(*args, **my_kwargs)
        self.depth_notes = {}

    def render(self, **kwargs):
        xp = kwargs.get("xp", None)
        if xp is True: # noqa this can also but something that evals to True but is not exactly true
            kwargs["xp"] = self.xp
        self.last_fig = self.renderer.render(self, **kwargs)
        return self.last_fig

    def CrossPlot(self, x=None, y=None, **kwargs):
        self.xp.reinit(x=x, y=y, **kwargs)  # TODO could add graph
        return self.xp

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    def set_render_setting(self, key, value):
        self._render[key] = value

    def show_depth(self, boolean):
        self.set_render_setting("show_depth", boolean)

    def set_depth_position(self, position):
        self.set_render_setting("depth_position", position)

    def set_height(self, height):
        self.set_render_setting("height", height)

    def set_depth(self, depth_range):
        self.set_render_setting("depth", depth_range)

    def get_render_settings(self):
        return self._render

    def process_data(self, *args, **kwargs):  # Add ways to add data
        for i, ar in enumerate(args):
            if str(type(ar)) == LAS_TYPE:
                self.add_las_object(ar, **kwargs)
            elif str(type(ar)) == WELLY_PROJECT_TYPE:
                if len(ar.get_wells()) != 1:
                    raise ValueError(
                        f"If you use welly, you must supply a well (or a project that has exactly one well). This project has {len(ar.get_wells())} wells."
                    )
                else:
                    self.add_welly_object(ar[0], **kwargs)
            elif str(type(ar)) == WELLY_WELL_TYPE:
                self.add_welly_object(ar, **kwargs)
            elif isinstance(ar, (pozo.Trace, pozo.Axis, pozo.Track)):
                self.add_tracks(ar)
            else:
                warnings.warn(
                    f"Unknown argument type passed: argument {i}, {type(ar)}. Ignored"
                )

    def add_las_object(self, ar, **kwargs):
        include = kwargs.get("include", [])
        exclude = kwargs.get("exclude", [])
        yaxis = kwargs.get("yaxis", None)
        yaxis_name = kwargs.get("yaxis_name", "DEPTH")
        yaxis_unit = None
        unit_map = kwargs.pop("unit_map", None)
        if (
            yaxis is not None
        ):  # this is a manually added y axis, don't parse it with LAS
            if yaxis_name is None and hasattr(yaxis, "mnemonic"):
                yaxis_name = yaxis.mnemonic
            if hasattr(yaxis, "unit"):
                yaxis_unit = yaxis.unit
            else:
                warnings.warn("Not sure what yaxis units are.")
            if hasattr(yaxis, "data"):
                yaxis = yaxis.data
            if len(yaxis) != len(ar.index):
                raise ValueError(
                    f"Length of supplied yaxis ({len(yaxis)}) does not match length of LAS File index ({len(ar.index)})"
                )
        elif yaxis_name in ar.curves.keys():
            yaxis = ar.curves[yaxis_name].data
            yaxis_unit = pzu.parse_unit_from_curve(ar.curves[yaxis_name])
        else:
            warnings.warn(
                "No yaxis specified and 'DEPTH' not found: using index. Set explicitly with yaxis= OR yaxis_name=. Not sure what y-axis units are."
            )
            yaxis = ar.depth_m
            yaxis_unit = pzu.registry.parse_units("meter")

        for curve in ar.curves:
            if yaxis_name is not None and curve.mnemonic == yaxis_name:
                continue

            mnemonic = pozo.deLASio(curve.mnemonic)
            if include and len(include) != 0 and curve.mnemonic not in include:
                continue
            elif exclude and len(exclude) != 0 and curve.mnemonic in exclude:
                continue
            unit = None
            if unit_map and (curve.mnemonic in unit_map or mnemonic in unit_map):
                if curve.mnemonic in unit_map:
                    unit = pzu.registry.parse_units(unit_map[curve.mnemonic])
                elif mnemonic in unit_map:
                    unit = pzu.registry.parse_units(unit_map[mnemonic])
            elif curve.unit is None:
                warnings.warn(
                    f"No units found for mnemonic {mnemonic}"
                )  # TODO Handle percentages/lookup mnemonics
            else:
                unit = pzu.parse_unit_from_curve(curve)

            if ooderr.NameConflictException(level=self._name_conflict) is None:
                name = mnemonic
            else:
                name = curve.mnemonic

            trace = pozo.Trace(
                curve.data,
                depth=yaxis,
                mnemonic=mnemonic,
                name=name,
                unit=unit,
                depth_unit=yaxis_unit,
            )
            self.add_tracks(trace)
        if include and len(include) != 0:
            self.reorder_all_tracks(include)

    def add_welly_object(self, ar, **kwargs):
        include = kwargs.get("include", [])
        exclude = kwargs.get("exclude", [])
        yaxis = kwargs.get("yaxis", None)
        yaxis_name = kwargs.get("yaxis_name", "DEPTH")
        yaxis_unit = kwargs.get("yaxis_unit", None)
        unit_map = kwargs.pop("unit_map", None)

        if yaxis is not None:
            if yaxis_name and hasattr(yaxis, "mnemonic"):
                yaxis_name = yaxis.mnemonic
            if hasattr(yaxis, "units"):
                yaxis_unit = pzu.registry.parse_unit_from_context(
                    pozo.deLASio(yaxis.mnemonic), yaxis.units, yaxis.values
                )
            else:
                warnings.warn("Not sure what yaxis units are.")  # TODO
            if hasattr(yaxis, "values"):
                yaxis = yaxis.values
        elif yaxis_name in ar.data.keys():
            yaxis = ar.data[yaxis_name]
            yaxis_unit = pzu.registry.parse_unit_from_context(
                pozo.deLASio(yaxis.mnemonic), yaxis.units, yaxis.values
            )
        else:
            raise ValueError(
                "No yaxis specified and 'DEPTH' not found. Set explicitly with yaxis= OR yaxis_name=."
            )
        for curve in ar.data.values():
            if yaxis_name is not None and curve.mnemonic == yaxis_name:
                continue

            mnemonic = pozo.deLASio(curve.mnemonic)
            if include and len(include) != 0 and curve.mnemonic not in include:
                continue
            elif exclude and len(exclude) != 0 and curve.mnemonic in exclude:
                continue

            unit = None

            if unit_map and (curve.mnemonic in unit_map or mnemonic in unit_map):
                if curve.mnemonic in unit_map:
                    unit = pzu.registry.parse_units(unit_map[curve.mnemonic])
                elif mnemonic in unit_map:
                    unit = pzu.registry.parse_units(unit_map[mnemonic])
            if curve.units is None:
                warnings.warn(f"No units found for mnemonic {mnemonic}")
            else:
                unit = pzu.registry.parse_unit_from_context(
                    mnemonic, curve.units, curve.values
                )  # TODO is curve correct?

            if ooderr.NameConflictException(level=self._name_conflict) is None:
                name = mnemonic
            else:
                name = curve.mnemonic
            depth = None
            depth_unit = None
            if yaxis is not None:
                depth = yaxis
                depth_unit = yaxis_unit
            else:
                depth = curve.index
                depth_unit = pzu.parse_unit_from_context(
                    pozo.deLASio(curve.index_name), curve.index_name, curve.index
                )

            trace = pozo.Trace(
                curve.values,
                depth=depth,
                mnemonic=mnemonic,
                name=name,
                unit=unit,
                depth_unit=depth_unit,
            )
            self.add_tracks(trace)
        if include and len(include) != 0:
            self.reorder_all_tracks(include)

    def _check_types(self, *tracks):
        accepted_types = (pozo.Axis, pozo.Trace, pozo.Track)
        raw_return = []
        for track in tracks:
            if isinstance(track, (list)):
                raw_return.extend(self._check_types(*tracks))
            elif not isinstance(track, accepted_types):
                raise TypeError(
                    "Axis.add_tracks() only accepts axes, tracks, and traces: pozo objects"
                )
            intermediate = track
            if isinstance(intermediate, pozo.Trace):
                intermediate = pozo.Axis(intermediate, name=intermediate.get_name())
            if isinstance(intermediate, pozo.Axis):
                intermediate = pozo.Track(intermediate, name=intermediate.get_name())
            if isinstance(intermediate, pozo.Track):
                raw_return.append(intermediate)
        return raw_return

    # add_items
    def add_axes(self, *axes, **kwargs):
        good_axes = self._check_types(*axes)
        super().add_items(*good_axes, **kwargs)
        return good_axes

    # add_items
    def add_tracks(self, *tracks, **kwargs):  # axis can take axes... and other axis?
        good_tracks = self._check_types(*tracks)
        super().add_items(*good_tracks, **kwargs)
        return good_tracks

    # get_items
    def get_tracks(self, *selectors, **kwargs):
        return super().get_items(*selectors, **kwargs)

    # get_item
    def get_track(self, selector, **kwargs):
        return super().get_item(selector, **kwargs)

    # pop items
    def pop_tracks(self, *selectors, **kwargs):
        return super().pop_items(*selectors, **kwargs)

    def combine_tracks(self, selector, *selectors):
        sink = self.get_track(selector, strict_index=False)
        if sink is None:
            if isinstance(selector, (pozo.Trace, pozo.Axis, pozo.Track)):
                self.add_tracks(selector)
                sink = selector
            else:
                raise TypeError(
                    "The first argument must be a track that exists or a new track track/axis/trace to add"
                )
        for sel in selectors:
            if not self.has_track(sel) and isinstance(sel, pozo.Track):
                self.add_tracks(sel)
        source = self.pop_tracks(*selectors, sort=False)
        for track in source:
            axes = track.pop_axes()
            sink.add_axes(axes)

    # what about whitelabelling all the other stuff
    def has_track(self, selector):
        return super().has_item(selector)

    def reorder_all_tracks(self, order):
        super().reorder_all_items(order)

    def move_tracks(self, *selectors, **kwargs):
        super().move_items(*selectors, **kwargs)

    def get_axes(self, *selectors, **kwargs):
        all_axes = []
        for track in self.get_tracks():
            all_axes.extend(track.get_axes(*selectors, **kwargs))
        return all_axes

    def get_traces(self, *selectors, **kwargs):
        all_traces = []
        for track in self.get_tracks():
            all_traces.extend(track.get_traces(*selectors, **kwargs))
        return all_traces

    def get_theme(self):
        context = {
            "type": "graph",
            "name": self._name,
        }
        return self._get_theme(context=context)



import ood

ood.exceptions.NameConflictException.default_level = ood.exceptions.ErrorLevel.IGNORE
ood.exceptions.MultiParentException.default_level = ood.exceptions.ErrorLevel.IGNORE
from .traces import Trace # noqa
from .axes import Axis # noqa
from .tracks import Track # noqa
from .graphs import Graph # noqa

import pozo.themes as themes # noqa
import pozo.renderers as renderers # noqa
import pozo.units as units # noqa

def deLASio(mnemonic):
    return mnemonic.split(":", 1)[0] if ":" in mnemonic else mnemonic

# Error on dimensionless
#


class HasLog(ood.selectors.Selector):
    def __init__(self, mnemonic):
        self.mnemonic = mnemonic
    def _process(self, parent):
        ret_items = []
        for item in parent.get_items():
            if hasattr(item, 'get_mnemonic'):
                if item.get_mnemonic() == self.mnemonic:
                    ret_items.append(item)
                    break
            if isinstance(item, ood.Observer):
                if item.has_item(self):
                    ret_items.append(item)
                    break
        return ret_items



import os
os.environ['PINT_ARRAY_PROTOCOL_FALLBACK'] = "0" # from numpy/pint documentation
import warnings
from io import StringIO

from IPython.core.display import HTML
import pandas as pd
import numpy as np
import pint

import pozo
class MissingRangeError(pint.UndefinedUnitError):
    pass

class LasMapEntry():
    def  __init__(self, boundries, unit, confidence):
        if not isinstance(boundries, tuple) or not (len(boundries) == 0 or len(boundries) == 2):
            raise TypeError("add_las_map ranges should contain a tuple with (min, max) or () catch-all")
        self.boundries = boundries
        self.unit = unit
        self.confidence = confidence

    def check(self, min_val, max_val):
        if len(self.boundries) == 0:
            return True
        elif min_val > self.boundries[0] and max_val < self.boundries[1]:
            return True
        return False

# Namespace would be nicer and I could hide registries if this wasn't overriden
# But would the map be global?
# Or would we just use a default registery like in init
class LasRegistry(pint.UnitRegistry):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mnemonic_units = {}
        self._reverse_mnemonic_units = {}

    def add_las_map(self, mnemonic, unit, ranges, confidence="- not indicated - LOW"):
        if not isinstance(ranges, list):
            if not isinstance(ranges, LasMapEntry):
                ranges = [LasMapEntry((), ranges, confidence)]
            else:
                ranges = [ranges]
        for ra in ranges:
            if not isinstance(ra, LasMapEntry):
                raise TypeError("Improperly formated map for LasMap, should be type LasMapEntry")
            self.parse_units(ra.unit) # Can't check in LasMapEntry because don't have registry
        if mnemonic not in self._mnemonic_units:
            self._mnemonic_units[mnemonic] = {}
            self._reverse_mnemonic_units[mnemonic] = {}
        self._mnemonic_units[mnemonic][unit] = ranges
        for ra in ranges:
            self._reverse_mnemonic_units[mnemonic][ra.unit] = unit

    def resolve_SI_unit_to_las(self, mnemonic, unit):
        # TODO: looking up the dimension would be nice
        mnemonic = pozo.deLASio(mnemonic)
        if mnemonic in self._reverse_mnemonic_units and unit in self._reverse_mnemonic_units[mnemonic]:
            return self._reverse_mnemonic_units[mnemonic][unit]
        else: return None


    def resolve_las_unit(self, mnemonic, unit, data):
        mnemonic = pozo.deLASio(mnemonic)
        # unit = unit
        max_val = np.nanmax(data)
        min_val = np.nanmin(data)
        if mnemonic in self._mnemonic_units and unit in self._mnemonic_units[mnemonic]:
            ranges = self._mnemonic_units[mnemonic][unit]
            for ra in ranges:
                if ra.check(min_val, max_val):
                    return ra
            raise MissingRangeError(f"{unit} for {mnemonic} found but not in range: {ranges}.")
        return None

    def parse_unit_from_context(self, mnemonic, unit, data): # this just gives you a result
        resolved = None
        try:
            resolved = self.resolve_las_unit(mnemonic, unit, data)
        except MissingRangeError as e:
            warnings.warn(str(e))
        if resolved is not None:
            return self.parse_units(resolved.unit)
        else:
            try:
                if not unit or unit == "": raise pint.UndefinedUnitError("Empty unit not allowed- please map it")
                return self.parse_units(unit)
            except pint.UndefinedUnitError as e:
                raise pint.UndefinedUnitError(f"{unit} for {pozo.deLASio(mnemonic)} not found. {str(e)}")

# we now don't get confidence from string



import warnings
import ood
import pozo
import pozo.renderers as pzr
import pozo.themes as pzt
import traceback

class Track(ood.Item, pzt.Themeable):
    _type="track"
    _child_type="axis"
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        for ar in args:
            self.add_axes(ar)

    def _check_types(self, *axes):
        accepted_types = (pozo.Axis, pozo.Trace)
        raw_return = []
        for axis in axes:
            if isinstance(axis, (list)):
                raw_return.extend(self._check_types(*axis))
            elif not isinstance(axis, accepted_types):
                raise TypeError(f"Axis.add_axes() only accepts axes, and traces: pozo objects, not {type(axis)}")
            elif isinstance(axis, pozo.Trace):
                raw_return.append(pozo.Axis(axis, name=axis.get_name()))
            else:
                raw_return.append(axis)
        return raw_return

    # add_items
    def add_axes(self, *axes, **kwargs):
        good_axes = self._check_types(*axes)
        super().add_items(*good_axes, **kwargs)
        return good_axes

    # get_items
    def get_axes(self, *selectors, **kwargs):
        return super().get_items(*selectors, **kwargs)

    # get_item
    def get_axis(self, selector, **kwargs):
        return super().get_item(selector, **kwargs)

    # pop items
    def pop_axes(self,  *selectors, **kwargs):
        return super().pop_items(*selectors, **kwargs)

    # what about whitelabelling all the other stuff
    def has_axis(self, selector):
        return super().has_item(selector)

    def reorder_all_axes(self, order):
        super().reorder_all_items(order)

    def move_axes(self, *selectors, **kwargs):
        super().move_items(*selectors, **kwargs)

    def get_traces(self, *selectors, **kwargs):
        ret_traces = []
        for axis in self.get_axes():
            ret_traces.extend(axis.get_traces(*selectors, **kwargs))
        return ret_traces

    def get_theme(self):
        context = { "type":"track",
                   "name": self._name,
                   }
        return self._get_theme(context=context)



import pytest
import ood
import ood.exceptions as e
import pozo.traces, pozo.axes

index = [1, 2, 3, 4]
values = [0, 1, 1, 0]
index2 = [5, 6, 7, 8]
values2 = [2, 3, 4, 5]

e.NameConflictException.default_level = e.ErrorLevel.IGNORE
e.MultiParentException.default_level = e.ErrorLevel.IGNORE

d1 = pozo.traces.Trace(values, depth=index, mnemonic="md1")
d2 = pozo.traces.Trace(values, depth=index, name="d2")
d3 = pozo.traces.Trace(values2, depth=index, name="d3", mnemonic="md3")
d4 = pozo.traces.Trace(values2, depth=index, mnemonic="md4")
d5 = pozo.traces.Trace(values, depth=index2, mnemonic="md5")
d6 = pozo.traces.Trace(values, depth=index2, mnemonic="md6")
d7 = pozo.traces.Trace(values2, depth=index2, mnemonic="md7")
d8 = pozo.traces.Trace(values2, depth=index2, mnemonic="md8")


def test_axis():
    a = pozo.axes.Axis()
    assert isinstance(a, pozo.axes.Axis)
    assert a.get_traces() == []
    with pytest.raises(ValueError):
        assert a.get_trace(None)
    assert a.pop_traces() == []
    b = pozo.axes.Axis(d1)
    assert len(b.get_traces()) ==1
    assert len(b) == 1
    assert b.get_trace(d1.get_name()) == d1
    assert b.get_trace(d1) == d1
    assert b.get_trace(d2) == None
    assert b.get_trace("md1") == d1
    assert b.get_trace('astoasi') == None
    assert b.get_trace(5) == None
    c = pozo.axes.Axis(*[d1, d2])
    d = pozo.axes.Axis(*[d1, d2], d3)
    e = pozo.axes.Axis(*[d4, d5], d6)
    assert e._count_dictionary() == 3
    assert len(e.get_traces()) == 3
    assert e._count_dictionary() == len(e.get_traces())
    assert e.get_traces() == [d4, d5, d6]
    assert e.get_trace(d4) == d4
    assert e.get_trace(d1) == None
    assert e.get_trace(d2) == None
    assert e.get_trace("md1") == None
    assert e.get_trace('astoasi') == None
    assert e.get_trace(5) == None
    assert e.get_trace("md4") == d4
    assert e.get_trace("md5") == d5
    assert e.get_trace("md6") == d6



    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_traces())
    with pytest.warns(UserWarning):
        e.add_traces(*[d5, d1]) # d1 new
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_traces())
    assert len(e.get_traces()) == 4
    with pytest.warns(UserWarning):
        e.add_traces(d1, d8, *[d2, d4]) #d2 d8 new
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_traces())
    assert len(e.get_traces()) == 6
    with pytest.warns(UserWarning):
        e.add_traces(*c.get_traces(), *d.get_traces(), *e.get_traces())
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_traces())
    assert len(e.get_traces()) == 7
    assert len(e.get_traces()) != 6
    e.add_traces()
    with pytest.raises(TypeError):
        e.add_traces(None)
        e.add_traces("whatever")
    assert e.get_traces("md6")[0] == d6
    assert e.get_traces("md6")[0] == d6
    assert e.get_traces("md6", "md7") == [d6]
    assert e.get_traces(d6, d7) == [d6]
    assert e.get_traces(d6) == [d6]
    n = len(e.get_traces())
    assert e.pop_traces(d6) == [d6]
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_traces())
    assert n == len(e.get_traces()) + 1
    # assert e.pop_traces(d6) == [] # pop is by default strict now
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_traces())
    #assert e.pop_traces(d6, d6, d6) == [] #by default strict
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_traces())
    e.pop_traces(*e.get_traces())
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_traces())
    assert len(e.get_traces()) == 0
    with pytest.warns(UserWarning):
        e.add_traces(*[d1,d2],*[d3,d4],*[d1,d4])
    assert e._count_dictionary() == len(e)
    assert e._count_dictionary() == len(e.get_traces())
    assert len(e.get_traces()) == 4
    assert e.get_traces(*[d1, d2, d3, d4]) == [d1, d2, d3, d4]
    assert e.get_traces(*[d1, d2],*[ d3, d4, d5]) == [d1, d2, d3, d4]
    assert e.get_traces(*["md1", d2, d3, d4]) == [d1, d2, d3, d4]
    assert e.get_trace(d1) == d1
    assert e.get_trace("md1") == d1
    d1.set_name("test")
    assert e.get_trace("test") == d1
    assert e.get_trace("md1") == None
